﻿/* Copyright 2010-2011 10gen Inc.
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
* http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace MongoDB.Bson.IO {
    /// <summary>
    /// Represents settings for a JsonWriter.
    /// </summary>
    [Serializable]
    public class JsonWriterSettings : BsonWriterSettings {
        #region private static fields
        private static JsonWriterSettings defaults = null; // delay creation to pick up the latest default values
        #endregion

        #region private fields
        private bool closeOutput = false;
        private Encoding encoding = Encoding.UTF8;
        private bool indent = false;
        private string indentChars = "  ";
        private string newLineChars = "\r\n";
        private JsonOutputMode outputMode = JsonOutputMode.Shell;
        private Version shellVersion = new Version(2, 0, 0);
        #endregion

        #region constructors
        /// <summary>
        /// Initializes a new instance of the JsonWriterSettings class.
        /// </summary>
        public JsonWriterSettings() {
        }
   
        /// <summary>
        /// Initializes a new instance of the JsonWriterSettings class.
        /// </summary>
        /// <param name="closeOutput">Whether to close the output when the writer is closed.</param>
        /// <param name="encoding">The output Encoding.</param>
        /// <param name="guidRepresentation">The representation for Guids.</param>
        /// <param name="indent">Whether to indent the output.</param>
        /// <param name="indentChars">The indentation characters.</param>
        /// <param name="newLineChars">The new line characters.</param>
        /// <param name="outputMode">The output mode.</param>
        /// <param name="shellVersion">The version of the shell to target.</param>
        public JsonWriterSettings(
            bool closeOutput,
            Encoding encoding,
            GuidRepresentation guidRepresentation,
            bool indent,
            string indentChars,
            string newLineChars,
            JsonOutputMode outputMode,
            Version shellVersion
        ) 
            : base(guidRepresentation) {
            this.closeOutput = closeOutput;
            this.encoding = encoding;
            this.indent = indent;
            this.indentChars = indentChars;
            this.newLineChars = newLineChars;
            this.outputMode = outputMode;
            this.shellVersion = shellVersion;
        }
        #endregion

        #region public static properties
        /// <summary>
        /// Gets or sets the default JsonWriterSettings.
        /// </summary>
        public static JsonWriterSettings Defaults {
            get {
                if (defaults == null) {
                    defaults = new JsonWriterSettings();
                }
                return defaults;
            }
            set { defaults = value; }
        }
        #endregion

        #region public properties
        /// <summary>
        /// Gets or sets whether to close the output when the writer is closed.
        /// </summary>
        public bool CloseOutput {
            get { return closeOutput; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                closeOutput = value;
            }
        }

        /// <summary>
        /// Gets or sets the output Encoding.
        /// </summary>
        public Encoding Encoding {
            get { return encoding; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                encoding = value;
            }
        }

        /// <summary>
        /// Gets or sets whether to indent the output.
        /// </summary>
        public bool Indent {
            get { return indent; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                indent = value;
            }
        }

        /// <summary>
        /// Gets or sets the indent characters.
        /// </summary>
        public string IndentChars {
            get { return indentChars; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                indentChars = value;
            }
        }

        /// <summary>
        /// Gets or sets the new line characters.
        /// </summary>
        public string NewLineChars {
            get { return newLineChars; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                newLineChars = value;
            }
        }

        /// <summary>
        /// Gets or sets the output mode.
        /// </summary>
        public JsonOutputMode OutputMode {
            get { return outputMode; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                outputMode = value;
            }
        }

        /// <summary>
        /// Gets or sets the shell version (used with OutputMode Shell).
        /// </summary>
        public Version ShellVersion {
            get { return shellVersion; }
            set {
                if (isFrozen) { throw new InvalidOperationException("JsonWriterSettings is frozen."); }
                shellVersion = value;
            }
        }
        #endregion

        #region public methods
        /// <summary>
        /// Creates a clone of the settings.
        /// </summary>
        /// <returns>A clone of the settings.</returns>
        public new JsonWriterSettings Clone() {
            return (JsonWriterSettings) CloneImplementation();
        }
        #endregion

        #region protected methods
        /// <summary>
        /// Creates a clone of the settings.
        /// </summary>
        /// <returns>A clone of the settings.</returns>
        protected override BsonWriterSettings CloneImplementation() {
            return new JsonWriterSettings(
                closeOutput,
                encoding,
                guidRepresentation,
                indent,
                indentChars,
                newLineChars,
                outputMode,
                shellVersion
            );
        }
        #endregion
    }
}
